from dataclasses import dataclass

from dateutil.relativedelta import relativedelta

from .base import Model


@dataclass
class ProcesType(Model):
    url: str
    nummer: int
    naam: str
    omschrijving: str
    toelichting: str
    procesobject: str


@dataclass
class Resultaat(Model):
    url: str
    proces_type: ProcesType
    nummer: int
    volledig_nummer: str
    generiek: bool
    specifiek: bool
    naam: str
    omschrijving: str
    herkomst: str
    waardering: str
    procestermijn: str
    procestermijn_weergave: str
    bewaartermijn: relativedelta
    toelichting: str



from itertools import groupby

from django import forms
from django.contrib.admin import widgets

from .service import get_informatieobjecttypen


def get_iot_choices():
    iots = get_informatieobjecttypen()
    iots = sorted(iots, key=lambda iot: (iot.catalogus.domein, iot.catalogus.rsin))

    for catalogus, _iots in groupby(iots, key=lambda iot: iot.catalogus):
        group = f"{catalogus.domein} - {catalogus.rsin}"
        iot_choices = [(iot.url, iot.omschrijving) for iot in _iots]
        yield (group, iot_choices)


class InformatieObjectTypeField(forms.ChoiceField):
    widget = widgets.AdminRadioSelect

    def __init__(self, choices=get_iot_choices, *args, **kwargs):
        super().__init__(choices=choices, *args, **kwargs)



from concurrent import futures
from typing import List, Optional, Type
from urllib.parse import parse_qs, urlparse

from .api_models.base import ZGWModel, factory
from .api_models.catalogi import Catalogus, InformatieObjectType
from .client import Client
from .constants import APITypes
from .models import Service


def get_paginated_results(
    client: Client,
    resource: str,
    minimum: Optional[int] = None,
    test_func: Optional[callable] = None,
    *args,
    **kwargs
) -> list:
    query_params = kwargs.get("query_params", {})

    results = []
    response = client.list(resource, *args, **kwargs)

    def _get_results(response):
        _results = response["results"]
        if test_func:
            _results = [result for result in _results if test_func(result)]
        return _results

    results += _get_results(response)

    if minimum and len(results) >= minimum:
        return results

    while response["next"]:
        next_url = urlparse(response["next"])
        query = parse_qs(next_url.query)
        new_page = int(query["page"][0])
        query_params["page"] = [new_page]
        kwargs["query_params"] = query_params
        response = client.list(resource, *args, **kwargs)
        results += _get_results(response)

        if minimum and len(results) >= minimum:
            return results

    return results


def _get_ztc_clients():
    services = Service.objects.filter(api_type=APITypes.ztc)
    clients = [service.build_client() for service in services]
    return clients


def _fetch_list(
    resource: str, clients: List[Client], model: Type[ZGWModel]
) -> List[ZGWModel]:
    def _fetch(client: Client):
        results = get_paginated_results(client, resource)
        return results

    with futures.ThreadPoolExecutor() as executor:
        resp_data = executor.map(_fetch, clients)
        flattened = sum(resp_data, [])

    return factory(model, flattened)


def get_catalogi(clients: List[Client] = None):
    if clients is None:
        clients = _get_ztc_clients()

    return _fetch_list("catalogus", clients, Catalogus)


def get_informatieobjecttypen(
    clients: List[Client] = None,
) -> List[InformatieObjectType]:
    """
    Retrieve all informatieobjecttypen for all catalogi.
    """
    if clients is None:
        clients = _get_ztc_clients()

    catalogi = {cat.url: cat for cat in get_catalogi(clients=clients)}
    iots = _fetch_list("informatieobjecttype", clients, InformatieObjectType)

    # resolve relations
    for iot in iots:
        iot.catalogus = catalogi[iot.catalogus]

    return iots



# Generated by Django 2.2.10 on 2020-03-27 11:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("zgw_consumers", "0006_update_oas_auth"),
    ]

    operations = [
        migrations.AddField(
            model_name="service",
            name="nlx",
            field=models.URLField(
                blank=True,
                help_text="NLX inway address",
                max_length=255,
                verbose_name="NLX url",
            ),
        ),
    ]



import uuid
from typing import Optional
from urllib.parse import urljoin, urlsplit, urlunsplit

from django.contrib.postgres.fields import JSONField
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models.functions import Length
from django.utils.translation import ugettext_lazy as _

from zds_client import ClientAuth

from .client import ZGWClient, get_client_class
from .constants import APITypes, AuthTypes
from .query import ServiceManager


class Service(models.Model):
    label = models.CharField(_("label"), max_length=100)
    api_type = models.CharField(_("type"), max_length=20, choices=APITypes.choices)
    api_root = models.CharField(_("api root url"), max_length=255, unique=True)

    extra = JSONField(
        _("extra configuration"),
        default=dict,
        help_text=_("Extra configuration that's service-specific"),
        blank=True,
    )

    # credentials for the API
    client_id = models.CharField(max_length=255, blank=True)
    secret = models.CharField(max_length=255, blank=True)
    auth_type = models.CharField(
        _("authorization type"), max_length=20, choices=AuthTypes, default=AuthTypes.zgw
    )
    header_key = models.CharField(_("header key"), max_length=100, blank=True)
    header_value = models.CharField(_("header value"), max_length=255, blank=True)
    oas = models.URLField(
        _("OAS"), max_length=1000, help_text=_("URL to OAS yaml file")
    )
    nlx = models.URLField(
        _("NLX url"), max_length=1000, blank=True, help_text=_("NLX (outway) address")
    )
    user_id = models.CharField(
        _("user ID"),
        max_length=255,
        blank=True,
        help_text=_(
            "User ID to use for the audit trail. Although these external API credentials are typically used by"
            "this API itself instead of a user, the user ID is required."
        ),
    )
    user_representation = models.CharField(
        _("user representation"),
        max_length=255,
        blank=True,
        help_text=_("Human readable representation of the user."),
    )

    objects = ServiceManager()

    class Meta:
        verbose_name = _("service")
        verbose_name_plural = _("services")

    def __str__(self):
        return f"[{self.get_api_type_display()}] {self.label}"

    def save(self, *args, **kwargs):
        if not self.api_root.endswith("/"):
            self.api_root = f"{self.api_root}/"

        if self.nlx and not self.nlx.endswith("/"):
            self.nlx = f"{self.nlx}/"

        if not self.oas:
            self.oas = urljoin(self.api_root, "schema/openapi.yaml")

        super().save(*args, **kwargs)

    def clean(self):
        super().clean()

        # validate extra for ztc service
        if self.api_type == APITypes.ztc:
            main_catalogus_uuid = self.extra.get("main_catalogus_uuid")
            if main_catalogus_uuid is None:
                raise ValidationError(
                    {"extra": _("Expected a 'main_catalogus_uuid' extra config")}
                )

            try:
                uuid.UUID(main_catalogus_uuid)
            except ValueError:
                raise ValidationError(
                    {
                        "extra": _(
                            "'main_catalogus_uuid' does not look like a valid UUID4"
                        )
                    }
                )

        # validate header_key and header_value
        if self.header_key and not self.header_value:
            raise ValidationError(
                {
                    "header_value": _(
                        "If header_key is set, header_value must also be set"
                    )
                }
            )
        if not self.header_key and self.header_value:
            raise ValidationError(
                {"header_key": _("If header_value is set, header_key must also be set")}
            )

    def build_client(self, **claims):
        """
        Build an API client from the service configuration.
        """
        _uuid = uuid.uuid4()

        api_root = self.api_root
        if self.nlx:
            api_root = api_root.replace(self.api_root, self.nlx, 1)

        dummy_detail_url = f"{api_root}dummy/{_uuid}"
        Client = get_client_class()
        client = Client.from_url(dummy_detail_url)
        client.schema_url = self.oas

        if self.auth_type == AuthTypes.zgw:
            client.auth = ClientAuth(
                client_id=self.client_id,
                secret=self.secret,
                user_id=self.user_id,
                user_representation=self.user_representation,
                **claims,
            )
        elif self.auth_type == AuthTypes.api_key:
            client.auth_value = {self.header_key: self.header_value}

        return client

    @classmethod
    def get_service(cls, url: str) -> "Service":
        split_url = urlsplit(url)
        scheme_and_domain = urlunsplit(split_url[:2] + ("", "", ""))

        candidates = (
            cls.objects.filter(api_root__startswith=scheme_and_domain)
            .annotate(api_root_length=Length("api_root"))
            .order_by("-api_root_length")
        )

        # select the one matching
        for candidate in candidates.iterator():
            if url.startswith(candidate.api_root):
                return candidate

        return None

    @classmethod
    def get_client(cls, url: str, **kwargs) -> Optional[ZGWClient]:
        service = cls.get_service(url)
        if not service:
            return None

        return service.build_client(**kwargs)

    @classmethod
    def get_auth_header(cls, url: str, **kwargs) -> Optional[dict]:
        client = cls.get_client(url, **kwargs)
        if not client:
            return None

        return client.auth_header



"""
Datamodels for ZGW resources.

These are NOT django models.
"""
import uuid
from datetime import date, datetime
from typing import Any, Dict, List, Union

from dateutil.parser import parse
from dateutil.relativedelta import relativedelta
from relativedeltafield import parse_relativedelta

from ._camel_case import underscoreize
from .types import JSONObject

__all__ = ["CONVERTERS", "factory", "Model", "ZGWModel"]


def noop(value: Any) -> Any:
    return value


CONVERTERS = {
    type(None): lambda x: None,
    str: noop,
    int: noop,
    float: noop,
    dict: noop,  # TODO: recurse?
    uuid.UUID: lambda value: uuid.UUID(value),
    datetime: parse,
    date: date.fromisoformat,
    relativedelta: parse_relativedelta,
}


class Model:
    def __post_init__(self):
        self._type_cast()

    def _type_cast(self):
        annotations = get_all_annotations(self.__class__)
        for attr, typehint in annotations.items():
            value = getattr(self, attr)

            if typehint is None:
                typehint = type(None)

            # support for Optional
            if hasattr(typehint, "__origin__") and typehint.__origin__ is Union:
                typehint = typehint.__args__

                if value is None:
                    continue

                # Optional is ONE type combined with None
                typehint = next(t for t in typehint if t is not None)

            if isinstance(value, typehint):
                continue

            converter = CONVERTERS[typehint]
            setattr(self, attr, converter(value))


class ZGWModel(Model):
    @property
    def uuid(self) -> uuid.UUID:
        """
        Because of the usage of UUID4, we can rely on the UUID as identifier.
        """
        _uuid = self.url.split("/")[-1]
        return uuid.UUID(_uuid)


def get_all_annotations(cls: type) -> Dict[str, Any]:
    annotations = {}
    for supercls in cls.__bases__:
        super_annotations = get_all_annotations(supercls)
        annotations.update(super_annotations)

    # Follow MRO - most specific top-level class wins, otherwise left-to-right
    if hasattr(cls, "__annotations__"):
        annotations.update(cls.__annotations__)
    return annotations


def factory(
    model: type, data: Union[JSONObject, List[JSONObject]]
) -> Union[type, List[type]]:
    _is_collection = isinstance(data, list)

    known_kwargs = list(get_all_annotations(model).keys())

    def _normalize(kwargs: dict):
        # TODO: this should be an explicit mapping, but *most* of the time with ZGW
        # API's this is fine.
        kwargs = underscoreize(kwargs)
        to_keep = {key: value for key, value in kwargs.items() if key in known_kwargs}
        return to_keep

    if not _is_collection:
        data = [data]

    instances = [model(**_normalize(_raw)) for _raw in data]

    if not _is_collection:
        instances = instances[0]
    return instances



import logging
from functools import lru_cache
from typing import Any, Dict, List, Tuple
from urllib.parse import parse_qs, urlparse

from django import forms
from django.contrib.admin import widgets
from django.db.models import Field
from django.http import HttpRequest

from .constants import APITypes
from .models import Service

logger = logging.getLogger(__name__)


# TODO: parallelize
@lru_cache()
def get_zaaktypen() -> Dict[Service, List[Dict[str, Any]]]:
    services = Service.objects.filter(api_type=APITypes.ztc)

    zaaktypen_per_service = {}

    for service in services:
        client = service.build_client()
        logger.debug("Fetching zaaktype list for service %r", service)
        zaaktypen_per_service[service] = []
        response = client.list(
            "zaaktype", catalogus_uuid=service.extra["main_catalogus_uuid"]
        )
        zaaktypen_per_service[service] += response["results"]
        while response["next"]:
            next_url = urlparse(response["next"])
            query = parse_qs(next_url.query)
            new_page = int(query["page"][0]) + 1
            query["page"] = [new_page]
            response = client.list(
                "zaaktype",
                catalogus_uuid=service.extra["main_catalogus_uuid"],
                query_params=query,
            )
            zaaktypen_per_service[service] += response["results"]

    return zaaktypen_per_service


def get_zaaktype_field(db_field: Field, request: HttpRequest, **kwargs):
    zaaktypen = get_zaaktypen()

    def _get_choice(zaaktype: dict) -> Tuple[str, str]:
        return (
            zaaktype["url"],
            f"{zaaktype['identificatie']} - {zaaktype['omschrijving']}",
        )

    choices = [
        (
            f"Service: {service.label}",
            [_get_choice(zaaktype) for zaaktype in _zaaktypen],
        )
        for service, _zaaktypen in zaaktypen.items()
    ]

    return forms.ChoiceField(
        label=db_field.verbose_name.capitalize(),
        widget=widgets.AdminRadioSelect(),
        choices=choices,
        required=False,
        help_text=db_field.help_text,
    )



import logging
from typing import Dict, Optional
from urllib.parse import urljoin

from django.conf import settings
from django.utils.module_loading import import_string

from zds_client import Client
from zds_client.oas import schema_fetcher

logger = logging.getLogger(__name__)


def get_client_class() -> type:
    client_class = getattr(
        settings, "ZGW_CONSUMERS_CLIENT_CLASS", "zgw_consumers.client.ZGWClient"
    )
    Client = import_string(client_class)
    return Client


class ZGWClient(Client):
    auth_value: Optional[Dict[str, str]] = None
    schema_url: str = ""

    def fetch_schema(self) -> None:
        """ support custom urls for OAS """
        url = self.schema_url or urljoin(self.base_url, "schema/openapi.yaml")
        logger.info("Fetching schema at '%s'", url)
        self._schema = schema_fetcher.fetch(url, {"v": "3"})

    def pre_request(self, method, url, **kwargs):
        """
        Add authorization header to requests for APIs without jwt.
        """
        if not self.auth and self.auth_value:
            headers = kwargs.get("headers", {})
            headers.update(self.auth_value)

        return super().pre_request(method, url, **kwargs)

    @property
    def auth_header(self) -> dict:
        if self.auth:
            return self.auth.credentials()

        return self.auth_value or {}


class UnknownService(Exception):
    pass



# Generated by Django 2.2.1 on 2019-05-14 08:09

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("zgw_consumers", "0002_auto_20190514_0944")]

    operations = [
        migrations.AlterField(
            model_name="service",
            name="api_type",
            field=models.CharField(
                choices=[
                    ("ac", "AC (Authorizations)"),
                    ("nrc", "NRC (Notifications)"),
                    ("zrc", "ZRC (Zaken)"),
                    ("ztc", "ZTC (Zaaktypen)"),
                    ("drc", "DRC (Informatieobjecten)"),
                    ("brc", "BRC (Besluiten)"),
                    ("orc", "ORC (Overige)"),
                ],
                max_length=20,
                verbose_name="type",
            ),
        )
    ]



from django.utils.translation import ugettext_lazy as _

from djchoices import ChoiceItem, DjangoChoices


class APITypes(DjangoChoices):
    ac = ChoiceItem("ac", _("AC (Authorizations)"))
    nrc = ChoiceItem("nrc", _("NRC (Notifications)"))

    zrc = ChoiceItem("zrc", _("ZRC (Zaken)"))
    ztc = ChoiceItem("ztc", _("ZTC (Zaaktypen)"))
    drc = ChoiceItem("drc", _("DRC (Informatieobjecten)"))
    brc = ChoiceItem("brc", _("BRC (Besluiten)"))
    kic = ChoiceItem("kic", _("KIC (Klantinteracties)"))

    orc = ChoiceItem("orc", _("ORC (Overige)"))


class AuthTypes(DjangoChoices):
    no_auth = ChoiceItem("no_auth", _("No authorization"))
    api_key = ChoiceItem("api_key", _("API key"))
    zgw = ChoiceItem("zgw", _("ZGW client_id + secret"))



from urllib.parse import urljoin

from django.db import migrations


def update_fields(apps, schema_editor):
    Service = apps.get_model("zgw_consumers", "Service")

    for service in Service.objects.filter(oas=""):
        service.oas = urljoin(service.api_root, "schema/openapi.yaml")
        service.save()


class Migration(migrations.Migration):
    dependencies = [
        ("zgw_consumers", "0005_auto_20200326_1040"),
    ]

    operations = [migrations.RunPython(update_fields, migrations.RunPython.noop)]



# Generated by Django 2.2.1 on 2019-05-14 07:44

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("zgw_consumers", "0001_initial")]

    operations = [
        migrations.AlterField(
            model_name="service",
            name="extra",
            field=django.contrib.postgres.fields.jsonb.JSONField(
                blank=True,
                default=dict,
                help_text="Extra configuration that's service-specific",
                verbose_name="extra configuration",
            ),
        )
    ]



# Generated by Django 2.2.10 on 2020-03-26 10:40

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("zgw_consumers", "0004_auto_20200113_1728"),
    ]

    operations = [
        migrations.AddField(
            model_name="service",
            name="auth_type",
            field=models.CharField(
                choices=[
                    ("no_auth", "No authorization"),
                    ("token", "Token"),
                    ("jwt", "JWT"),
                ],
                default="jwt",
                max_length=20,
                verbose_name="authorization type",
            ),
        ),
        migrations.AddField(
            model_name="service",
            name="header_key",
            field=models.CharField(
                blank=True, max_length=100, verbose_name="header key"
            ),
        ),
        migrations.AddField(
            model_name="service",
            name="header_value",
            field=models.CharField(
                blank=True, max_length=255, verbose_name="header value"
            ),
        ),
        migrations.AddField(
            model_name="service",
            name="oas",
            field=models.URLField(
                default="",
                help_text="Path to OAS yaml file",
                max_length=255,
                verbose_name="OAS",
            ),
            preserve_default=False,
        ),
    ]



# Source: https://github.com/vbabiy/djangorestframework-camel-case/blob/master/djangorestframework_camel_case/util.py

# Copyright (c) 2013, Vitaly Babiy
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Django REST Framework JSON CamelCase nor the names of
#   its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import re

from django.core.files import File
from django.http import QueryDict


def get_underscoreize_re(options):
    if options.get("no_underscore_before_number"):
        pattern = r"([a-z]|[0-9]+[a-z]?|[A-Z]?)([A-Z])"
    else:
        pattern = r"([a-z]|[0-9]+[a-z]?|[A-Z]?)([A-Z0-9])"
    return re.compile(pattern)


def camel_to_underscore(name, **options):
    underscoreize_re = get_underscoreize_re(options)
    return underscoreize_re.sub(r"\1_\2", name).lower()


def _get_iterable(data):
    if isinstance(data, QueryDict):
        return data.lists()
    else:
        return data.items()


def underscoreize(data, **options):
    if isinstance(data, dict):
        new_dict = {}
        for key, value in _get_iterable(data):
            if isinstance(key, str):
                new_key = camel_to_underscore(key, **options)
            else:
                new_key = key
            new_dict[new_key] = underscoreize(value, **options)

        if isinstance(data, QueryDict):
            new_query = QueryDict(mutable=True)
            for key, value in new_dict.items():
                new_query.setlist(key, value)
            return new_query
        return new_dict
    if is_iterable(data) and not isinstance(data, (str, File)):
        return [underscoreize(item, **options) for item in data]

    return data


def is_iterable(obj):
    try:
        iter(obj)
    except TypeError:
        return False
    else:
        return True



from dataclasses import dataclass
from datetime import date
from typing import Optional

from .base import ZGWModel
from .constants import VertrouwelijkheidsAanduidingen


@dataclass
class Document(ZGWModel):
    url: str
    auteur: str
    identificatie: str
    beschrijving: str
    bestandsnaam: str
    bestandsomvang: int
    bronorganisatie: str
    creatiedatum: date
    formaat: str  # noqa
    indicatie_gebruiksrecht: dict
    informatieobjecttype: str
    inhoud: str
    integriteit: dict
    link: str
    ondertekening: dict
    ontvangstdatum: Optional[date]
    status: str
    taal: str
    titel: str
    vertrouwelijkheidaanduiding: str
    verzenddatum: Optional[date]

    def get_vertrouwelijkheidaanduiding_display(self):
        return VertrouwelijkheidsAanduidingen.values[self.vertrouwelijkheidaanduiding]



# Generated by Django 2.2.1 on 2019-05-14 07:40

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Service",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("label", models.CharField(max_length=100, verbose_name="label")),
                (
                    "api_type",
                    models.CharField(
                        choices=[
                            ("ac", "AC (Authorizations)"),
                            ("nrc", "NRC (Notifications)"),
                            ("zrc", "ZRC (Zaken)"),
                            ("ztc", "ZTC (Zaaktypen)"),
                            ("drc", "DRC (Informatieobjecten"),
                            ("brc", "BRC (Besluiten)"),
                            ("orc", "ORC (Overige)"),
                        ],
                        max_length=20,
                        verbose_name="type",
                    ),
                ),
                (
                    "api_root",
                    models.CharField(
                        max_length=255, unique=True, verbose_name="api root url"
                    ),
                ),
                (
                    "extra",
                    django.contrib.postgres.fields.jsonb.JSONField(
                        default=dict,
                        help_text="Extra configuration that's service-specific",
                        verbose_name="extra configuration",
                    ),
                ),
                ("client_id", models.CharField(blank=True, max_length=255)),
                ("secret", models.CharField(blank=True, max_length=255)),
            ],
            options={"verbose_name": "service", "verbose_name_plural": "services"},
        )
    ]



from dataclasses import dataclass
from datetime import date, datetime
from typing import Any, Optional

from .base import ZGWModel
from .catalogi import Eigenschap
from .constants import RolOmschrijving, RolTypes, VertrouwelijkheidsAanduidingen


@dataclass
class Zaak(ZGWModel):
    url: str
    identificatie: str
    bronorganisatie: str
    omschrijving: str
    toelichting: str
    zaaktype: str
    registratiedatum: date
    startdatum: date
    einddatum: Optional[date]
    einddatum_gepland: Optional[date]
    uiterlijke_einddatum_afdoening: Optional[date]
    publicatiedatum: Optional[date]
    vertrouwelijkheidaanduiding: str
    status: str
    resultaat: str
    relevante_andere_zaken: list
    zaakgeometrie: dict

    def get_vertrouwelijkheidaanduiding_display(self):
        return VertrouwelijkheidsAanduidingen.values[self.vertrouwelijkheidaanduiding]


@dataclass
class Status(ZGWModel):
    url: str
    zaak: str
    statustype: str
    datum_status_gezet: datetime
    statustoelichting: str


@dataclass
class ZaakObject(ZGWModel):
    url: str
    zaak: str
    object: str
    object_type: str
    object_type_overige: str
    relatieomschrijving: str


@dataclass
class ZaakEigenschap(ZGWModel):
    url: str
    # uuid: uuid.UUID
    zaak: str
    eigenschap: str
    naam: str
    waarde: str

    def get_waarde(self) -> Any:
        assert isinstance(
            self.eigenschap, Eigenschap
        ), "Ensure eigenschap has been resolved"
        return self.eigenschap.to_python(self.waarde)


@dataclass
class Resultaat(ZGWModel):
    url: str
    zaak: str
    resultaattype: str
    toelichting: str


@dataclass
class Rol(ZGWModel):
    url: str
    zaak: str
    betrokkene: str
    betrokkene_type: str
    roltype: str
    omschrijving: str
    omschrijving_generiek: str
    roltoelichting: str
    registratiedatum: datetime
    indicatie_machtiging: str
    betrokkene_identificatie: Optional[dict]

    def get_betrokkene_type_display(self):
        return RolTypes.values[self.betrokkene_type]

    def get_omschrijving_generiek_display(self):
        return RolOmschrijving.values[self.omschrijving_generiek]



# Generated by Django 2.2.9 on 2020-01-13 16:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("zgw_consumers", "0003_auto_20190514_1009"),
    ]

    operations = [
        migrations.AlterField(
            model_name="service",
            name="api_type",
            field=models.CharField(
                choices=[
                    ("ac", "AC (Authorizations)"),
                    ("nrc", "NRC (Notifications)"),
                    ("zrc", "ZRC (Zaken)"),
                    ("ztc", "ZTC (Zaaktypen)"),
                    ("drc", "DRC (Informatieobjecten)"),
                    ("brc", "BRC (Besluiten)"),
                    ("kic", "KIC (Klantinteracties)"),
                    ("orc", "ORC (Overige)"),
                ],
                max_length=20,
                verbose_name="type",
            ),
        ),
    ]



# Generated by Django 2.2.10 on 2020-03-31 14:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("zgw_consumers", "0007_service_nlx"),
    ]

    operations = [
        migrations.AlterField(
            model_name="service",
            name="auth_type",
            field=models.CharField(
                choices=[
                    ("no_auth", "No authorization"),
                    ("api_key", "API key"),
                    ("zgw", "ZGW client_id + secret"),
                ],
                default="zgw",
                max_length=20,
                verbose_name="authorization type",
            ),
        ),
        migrations.AlterField(
            model_name="service",
            name="nlx",
            field=models.URLField(
                blank=True,
                help_text="NLX (outway) address",
                max_length=1000,
                verbose_name="NLX url",
            ),
        ),
        migrations.AlterField(
            model_name="service",
            name="oas",
            field=models.URLField(
                help_text="URL to OAS yaml file", max_length=1000, verbose_name="OAS"
            ),
        ),
    ]



# Generated by Django 2.2.10 on 2020-04-01 08:29

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("zgw_consumers", "0008_auto_20200331_1400")]

    operations = [
        migrations.AddField(
            model_name="service",
            name="user_id",
            field=models.CharField(
                blank=True,
                help_text="User ID to use for the audit trail. Although these external API credentials are typically used bythis API itself instead of a user, the user ID is required.",
                max_length=255,
                verbose_name="user ID",
            ),
        ),
        migrations.AddField(
            model_name="service",
            name="user_representation",
            field=models.CharField(
                blank=True,
                help_text="Human readable representation of the user.",
                max_length=255,
                verbose_name="user representation",
            ),
        ),
    ]



from dataclasses import dataclass
from datetime import date, datetime
from decimal import Decimal
from typing import Optional, Union

from dateutil.parser import parse
from dateutil.relativedelta import relativedelta

from .base import Model, ZGWModel, factory


@dataclass
class Catalogus(ZGWModel):
    url: str
    domein: str
    rsin: str
    contactpersoon_beheer_emailadres: str
    contactpersoon_beheer_naam: str
    contactpersoon_beheer_telefoonnummer: str
    besluittypen: list
    informatieobjecttypen: list
    zaaktypen: list


@dataclass
class ZaakType(ZGWModel):
    url: str
    catalogus: str
    identificatie: int
    omschrijving: str
    omschrijving_generiek: str
    vertrouwelijkheidaanduiding: str
    doel: str
    aanleiding: str
    toelichting: str
    indicatie_intern_of_extern: str
    handeling_initiator: str
    onderwerp: str
    handeling_behandelaar: str
    doorlooptijd: relativedelta
    servicenorm: Optional[relativedelta]
    opschorting_en_aanhouding_mogelijk: bool
    verlenging_mogelijk: bool
    verlengingstermijn: Optional[relativedelta]
    trefwoorden: list
    publicatie_indicatie: bool
    publicatietekst: str
    verantwoordingsrelatie: list
    producten_of_diensten: list
    # selectielijst_procestype: ProcesType
    statustypen: list
    resultaattypen: list
    eigenschappen: list
    informatieobjecttypen: list
    roltypen: list
    besluittypen: list
    deelzaaktypen: list

    begin_geldigheid: date
    einde_geldigheid: Optional[date]
    versiedatum: date
    concept: bool


@dataclass
class StatusType(ZGWModel):
    url: str
    zaaktype: str
    omschrijving: str
    omschrijving_generiek: str
    statustekst: str
    volgnummer: int
    is_eindstatus: bool


@dataclass
class InformatieObjectType(ZGWModel):
    url: str
    catalogus: str
    omschrijving: str
    vertrouwelijkheidaanduiding: str
    begin_geldigheid: date
    einde_geldigheid: Optional[date]
    concept: bool


@dataclass
class ResultaatType(ZGWModel):
    url: str
    zaaktype: str
    omschrijving: str
    resultaattypeomschrijving: str
    omschrijving_generiek: str
    selectielijstklasse: str
    toelichting: str
    archiefnominatie: str
    archiefactietermijn: Optional[relativedelta]
    brondatum_archiefprocedure: dict


@dataclass
class EigenschapSpecificatie(Model):
    groep: str
    formaat: str
    lengte: str
    kardinaliteit: str
    waardenverzameling: list  # List[str]


EIGENSCHAP_FORMATEN = {
    "tekst": str,
    "getal": lambda val: Decimal(val.replace(",", ".")),
    "datum": date.fromisoformat,
    "datum_type": parse,
}


@dataclass
class Eigenschap(ZGWModel):
    url: str
    zaaktype: str
    naam: str
    definitie: str
    specificatie: dict
    toelichting: str

    def __post_init__(self):
        super().__post_init__()
        self.specificatie = factory(EigenschapSpecificatie, self.specificatie)

    def to_python(self, value: str) -> Union[str, Decimal, date, datetime]:
        """
        Cast the string value into the appropriate python type based on the spec.
        """
        formaat = self.specificatie.formaat
        assert formaat in EIGENSCHAP_FORMATEN, f"Unknown format {formaat}"

        converter = EIGENSCHAP_FORMATEN[formaat]
        return converter(value)


@dataclass
class RolType(ZGWModel):
    url: str
    zaaktype: str
    omschrijving: str
    omschrijving_generiek: str



"""
Rewrite the URLs in anything that looks like a string, dict or list.
"""
from typing import Any, Iterable, List, Optional, Union

from zds_client.client import Object

from .models import Service


def _rewrite_url(value: str, rewrites: Iterable) -> Optional[str]:
    for start, replacement in rewrites:
        if not value.startswith(start):
            continue

        return value.replace(start, replacement, 1)

    return None


class Rewriter:
    def __init__(self):
        self.rewrites = Service.objects.exclude(nlx="").values_list("api_root", "nlx")

    @property
    def reverse_rewrites(self):
        return [(to_value, from_value) for from_value, to_value in self.rewrites]

    def forwards(self, data: Union[list, dict]):
        """
        Rewrite URLs from from_value to to_value.
        """
        self._rewrite(data, self.rewrites)

    def backwards(self, data: Union[list, dict]):
        """
        Rewrite URLs from to_value to from_value.
        """
        self._rewrite(data, self.reverse_rewrites)

    def _rewrite(self, data: Union[list, dict], rewrites: Iterable) -> None:
        if isinstance(data, list):
            new_items = []
            for item in data:
                if isinstance(item, str):
                    new_value = _rewrite_url(item, rewrites)
                    if new_value:
                        new_items.append(new_value)
                    else:
                        new_items.append(item)
                else:
                    self._rewrite(item, rewrites=rewrites)
                    new_items.append(item)

            # replace list elements
            assert len(new_items) == len(data)
            for i in range(len(data)):
                data[i] = new_items[i]
            return

        if not isinstance(data, dict):
            return

        for key, value in data.items():
            if isinstance(value, (dict, list)):
                self._rewrite(value, rewrites=rewrites)
                continue

            elif not isinstance(value, str):
                continue

            assert isinstance(value, str)

            rewritten = _rewrite_url(value, rewrites)
            if rewritten is not None:
                data[key] = rewritten


class NLXClientMixin:
    """
    Enable URL rewriting for zds_client.Client clients.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.rewriter = Rewriter()

    # def pre_request(self, method: str, url: str, **kwargs) -> Any:
    #     """
    #     Rewrite NLX urls in the request body and params.

    #     From NLX -> canonical.
    #     """
    #     json = kwargs.get("json")
    #     if json:
    #         self.rewriter.backwards(json)

    #     params = kwargs.get("params")
    #     if params:
    #         self.rewriter.backwards(params)

    #     return super().pre_request(method, url, **kwargs)

    def request(
        self, path: str, operation: str, method="GET", expected_status=200, **kwargs
    ) -> Union[List[Object], Object]:
        """
        Make the actual HTTP request.
        """
        # intercept canonical URLs and rewrite to NLX
        _paths = [path]
        self.rewriter.forwards(_paths)
        path = _paths[0]

        return super().request(
            path, operation, method=method, expected_status=expected_status, **kwargs
        )

    def post_response(
        self, pre_id: Any, response_data: Optional[Union[dict, list]] = None
    ) -> None:
        """
        Rewrite from NLX -> canonical.
        """
        if response_data:
            self.rewriter.backwards(response_data)
        super().post_response(pre_id, response_data)



from djchoices import ChoiceItem, DjangoChoices


class VertrouwelijkheidsAanduidingen(DjangoChoices):
    openbaar = ChoiceItem("openbaar", label="Openbaar")
    beperkt_openbaar = ChoiceItem("beperkt_openbaar", label="Beperkt openbaar")
    intern = ChoiceItem("intern", label="Intern")
    zaakvertrouwelijk = ChoiceItem("zaakvertrouwelijk", label="Zaakvertrouwelijk")
    vertrouwelijk = ChoiceItem("vertrouwelijk", label="Vertrouwelijk")
    confidentieel = ChoiceItem("confidentieel", label="Confidentieel")
    geheim = ChoiceItem("geheim", label="Geheim")
    zeer_geheim = ChoiceItem("zeer_geheim", label="Zeer geheim")


class RolTypes(DjangoChoices):
    natuurlijk_persoon = ChoiceItem("natuurlijk_persoon", "Natuurlijk persoon")
    niet_natuurlijk_persoon = ChoiceItem(
        "niet_natuurlijk_persoon", "Niet-natuurlijk persoon"
    )
    vestiging = ChoiceItem("vestiging", "Vestiging")
    organisatorische_eenheid = ChoiceItem(
        "organisatorische_eenheid", "Organisatorische eenheid"
    )
    medewerker = ChoiceItem("medewerker", "Medewerker")


class RolOmschrijving(DjangoChoices):
    adviseur = ChoiceItem(
        "adviseur",
        "Adviseur",
        description="Kennis in dienst stellen van de behandeling van (een deel van) een zaak.",
    )
    behandelaar = ChoiceItem(
        "behandelaar",
        "Behandelaar",
        description="De vakinhoudelijke behandeling doen van (een deel van) een zaak.",
    )
    belanghebbende = ChoiceItem(
        "belanghebbende",
        "Belanghebbende",
        description="Vanuit eigen en objectief belang rechtstreeks betrokken "
        "zijn bij de behandeling en/of de uitkomst van een zaak.",
    )
    beslisser = ChoiceItem(
        "beslisser",
        "Beslisser",
        description="Nemen van besluiten die voor de uitkomst van een zaak noodzakelijk zijn.",
    )
    initiator = ChoiceItem(
        "initiator",
        "Initiator",
        description="Aanleiding geven tot de start van een zaak ..",
    )
    klantcontacter = ChoiceItem(
        "klantcontacter",
        "Klantcontacter",
        description="Het eerste aanspreekpunt zijn voor vragen van burgers en bedrijven ..",
    )
    zaakcoordinator = ChoiceItem(
        "zaakcoordinator",
        "Zaakcoördinator",
        description="Er voor zorg dragen dat de behandeling van de zaak in samenhang "
        "uitgevoerd wordt conform de daarover gemaakte afspraken.",
    )
    medeinitiator = ChoiceItem("mede_initiator", "Mede-initiator", description="")



from urllib.parse import urlsplit, urlunsplit

from django.db import models
from django.db.models.functions import Length

from zds_client import Client


class ServiceManager(models.Manager):
    def get_client_for(self, url: str) -> Client:
        """
        Return a Client instance capable of fetching the url.

        The URL is supposed to represent a resource in the API. Based on the URL, the
        service object is returned that can be used to fetch the object.
        """
        split_url = urlsplit(url)
        scheme_and_domain = urlunsplit(split_url[:2] + ("", "", ""))

        candidates = (
            self.filter(api_root__startswith=scheme_and_domain)
            .annotate(api_root_length=Length("api_root"))
            .order_by("-api_root_length")
        )

        # select the one matching
        for candidate in candidates.iterator():
            if url.startswith(candidate.api_root):
                service = candidate
                break
        else:
            raise self.model.DoesNotExist(f"No service found for url '{url}'")

        return service.build_client()



from django.contrib import admin

from .admin_fields import get_zaaktype_field
from .models import Service


@admin.register(Service)
class ServiceAdmin(admin.ModelAdmin):
    list_display = ("label", "api_type", "api_root", "auth_type")
    list_filter = ("api_type", "auth_type")
    search_fields = ("label", "api_root", "nlx")


class ListZaaktypenMixin:
    zaaktype_fields = ()

    def formfield_for_dbfield(self, db_field, request, **kwargs):
        if db_field.name in self.zaaktype_fields:
            return get_zaaktype_field(db_field, request, **kwargs)
        return super().formfield_for_dbfield(db_field, request, **kwargs)
